@@ -1,7 +1,7 @@
 import math
 from typing import Optional
 
-from .transformer import TransformerEncoder, DenoisingAutoEncoder
+from .transformer import TransformerEncoder, FeedforwardLayer
 from .cnn import SpaceToDepth, calculate_output_spatial_size, spatial_tuple
 from .activation import ReLU, SquaredReLU, GELU, SwiGLU
 from einops import einsum
@@ -27,7 +27,7 @@ class SequencePool(nn.Module):
         as a generalisation of average pooling.
     """
 
-    def __init__(self, d_model, linear_module, out_dim):
+    def __init__(self, d_model, linear_module, out_dim, batch_norm=True):
         super().__init__()
         self.d_model = d_model
         self.attention = nn.Sequential(
@@ -38,7 +38,11 @@ class SequencePool(nn.Module):
             ]
         )
         self.projection = nn.Linear(d_model, out_dim)
-        self.norm = nn.BatchNorm1d(out_dim, affine=False)
+        self.batch_norm = batch_norm
+        if batch_norm:
+            self.norm = nn.BatchNorm1d(out_dim, affine=False)
+        else:
+            self.norm = None
 
     def forward(self, x):
         weights = self.attention(x)
@@ -46,10 +50,10 @@ class SequencePool(nn.Module):
             weights, x, "batch seq, batch seq d_model -> batch d_model"
         )
         projection = self.projection(weighted_embedding)
-        return self.norm(projection)
+        return self.norm(projection) if self.batch_norm else projection
 
 
-class CCTEncoder(nn.Module):
+class ViTEncoder(nn.Module):
     """
     Based on the Compact Convolutional Transformer (CCT) of [Hasani et al. (2021)
         *''Escaping the Big Data Paradigm with Compact Transformers''*](
@@ -66,29 +70,30 @@ class CCTEncoder(nn.Module):
         minimum_cnn_out_channels=16,
         cnn_kernel_size=3,
         cnn_kernel_stride=1,
-        cnn_kernel_padding="same",
+        cnn_padding="same",
         cnn_kernel_dilation=1,
         cnn_kernel_groups=1,
-        cnn_activation: nn.Module = nn.ReLU,
+        cnn_activation: nn.Module = ReLU,
         cnn_activation_kwargs: Optional[dict] = None,
         cnn_dropout=0.0,
-        pooling_type="maxpool",
+        pooling_type="concat",  # max, average or concat
         pooling_kernel_size=3,
         pooling_kernel_stride=2,
-        pooling_kernel_padding=1,
-        transformer_position_embedding="absolute",  # absolute or relative
+        pooling_padding=1,
+        intermediate_feedforward_layer=True,
+        transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
         transformer_mlp_ratio=2,
-        transformer_bos_tokens=4,
-        transformer_activation: nn.Module = nn.GELU,
+        transformer_bos_tokens=0,
+        transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
-        mlp_dropout=0.0,
-        msa_dropout=0.1,
-        stochastic_depth=0.1,
+        transformer_mlp_dropout=0.0,
+        transformer_msa_dropout=0.1,
+        transformer_stochastic_depth=0.1,
         linear_module=nn.Linear,
-        batch_norm=True,
+        initial_batch_norm=True,
     ):
         super().__init__()
 
@@ -109,16 +114,19 @@ class CCTEncoder(nn.Module):
 
         if self.spatial_dimensions == 1:
             maxpoolxd = nn.MaxPool1d
+            avgpoolxd = nn.AvgPool1d
             convxd = nn.Conv1d
             batchnormxd = nn.BatchNorm1d
             spatial_dim_names = "D1"
         elif self.spatial_dimensions == 2:
             maxpoolxd = nn.MaxPool2d
+            avgpoolxd = nn.AvgPool2d
             convxd = nn.Conv2d
             batchnormxd = nn.BatchNorm2d
             spatial_dim_names = "D1 D2"
         elif self.spatial_dimensions == 3:
             maxpoolxd = nn.MaxPool3d
+            avgpoolxd = nn.AvgPool3d
             convxd = nn.Conv3d
             batchnormxd = nn.BatchNorm3d
             spatial_dim_names = "D1 D2 D3"
@@ -131,7 +139,7 @@ class CCTEncoder(nn.Module):
             input_size,
             kernel_size=cnn_kernel_size,
             stride=cnn_kernel_stride,
-            padding=cnn_kernel_padding,
+            padding=cnn_padding,
             dilation=cnn_kernel_dilation,
         )
 
@@ -142,7 +150,7 @@ class CCTEncoder(nn.Module):
                 cnn_output_size,
                 kernel_size=pooling_kernel_size,
                 stride=pooling_kernel_stride,
-                padding=pooling_kernel_padding,
+                padding=pooling_padding,
                 dilation=1,
             )
         )
@@ -153,15 +161,17 @@ class CCTEncoder(nn.Module):
             spatial_tuple(pooling_kernel_size, self.spatial_dimensions)
         )
 
-        if pooling_type in ["maxpool", None]:
+        if pooling_type in ["max", "average", None]:
             cnn_out_channels = transformer_embedding_size
         elif pooling_type == "concat":
-            cnn_out_channels = min(
+            cnn_out_channels = max(
                 math.floor(transformer_embedding_size / pooling_kernel_voxels),
                 minimum_cnn_out_channels,
             )
         else:
-            raise NotImplementedError("Pooling type must be maxpool, concat or None")
+            raise NotImplementedError(
+                "Pooling type must be max, average, concat or None"
+            )
 
         cnn_activation_out_channels = cnn_out_channels
 
@@ -174,7 +184,7 @@ class CCTEncoder(nn.Module):
             cnn_out_channels,
             cnn_kernel_size,
             stride=cnn_kernel_stride,
-            padding=cnn_kernel_padding,
+            padding=cnn_padding,
             dilation=cnn_kernel_dilation,
             groups=cnn_kernel_groups,
             bias=True,
@@ -191,7 +201,7 @@ class CCTEncoder(nn.Module):
                 nn.Dropout(cnn_dropout),
                 (
                     batchnormxd(cnn_activation_out_channels)
-                    if batch_norm
+                    if initial_batch_norm
                     else nn.Identity()
                 ),
             ]
@@ -206,13 +216,27 @@ class CCTEncoder(nn.Module):
                 ]
             )
 
-        elif pooling_type == "maxpool":
+        elif pooling_type == "max":
             self.pool = nn.Sequential(
                 *[
                     maxpoolxd(
                         pooling_kernel_size,
                         stride=pooling_kernel_stride,
-                        padding=pooling_kernel_padding,
+                        padding=pooling_padding,
+                    ),
+                    Rearrange(
+                        f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
+                    ),  # for transformer
+                ]
+            )
+
+        elif pooling_type == "average":
+            self.pool = nn.Sequential(
+                *[
+                    avgpoolxd(
+                        pooling_kernel_size,
+                        stride=pooling_kernel_stride,
+                        padding=pooling_padding,
                     ),
                     Rearrange(
                         f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
@@ -238,20 +262,14 @@ class CCTEncoder(nn.Module):
                     SpaceToDepth(
                         pooling_kernel_size,
                         stride=pooling_kernel_stride,
-                        padding=pooling_kernel_padding,
+                        padding=pooling_padding,
                         spatial_dimensions=self.spatial_dimensions,
                     ),
                     Rearrange(  # for transformer
                         f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
                     ),
-                    DenoisingAutoEncoder(
-                        concatpool_out_channels,
-                        transformer_mlp_ratio,
-                        transformer_embedding_size,
-                        activation=transformer_activation,
-                        activation_kwargs=transformer_activation_kwargs,
-                        dropout=0.0,
-                        linear_module=linear_module,
+                    PadTensor(
+                        (0, transformer_embedding_size - concatpool_out_channels)
                     ),
                 ]
             )
@@ -267,9 +285,9 @@ class CCTEncoder(nn.Module):
                 mlp_ratio=transformer_mlp_ratio,
                 activation=transformer_activation,
                 activation_kwargs=transformer_activation_kwargs,
-                mlp_dropout=mlp_dropout,
-                msa_dropout=msa_dropout,
-                stochastic_depth=stochastic_depth,
+                mlp_dropout=transformer_mlp_dropout,
+                msa_dropout=transformer_msa_dropout,
+                stochastic_depth=transformer_stochastic_depth,
                 causal=False,
                 linear_module=linear_module,
                 bos_tokens=transformer_bos_tokens,
@@ -279,10 +297,23 @@ class CCTEncoder(nn.Module):
 
         self.encoder = nn.Sequential(
             *[
-                batchnormxd(cnn_in_channels) if batch_norm else nn.Identity(),
+                batchnormxd(cnn_in_channels) if initial_batch_norm else nn.Identity(),
                 self.cnn,
                 self.activate_and_dropout,
                 self.pool,
+                (
+                    FeedforwardLayer(
+                        transformer_embedding_size,
+                        transformer_mlp_ratio,
+                        transformer_embedding_size,
+                        activation=transformer_activation,
+                        activation_kwargs=transformer_activation_kwargs,
+                        dropout=transformer_mlp_dropout,
+                        linear_module=linear_module,
+                    )
+                    if intermediate_feedforward_layer
+                    else nn.Identity()
+                ),
                 self.transformer,
             ]
         )
@@ -293,6 +324,7 @@ class CCTEncoder(nn.Module):
 
 class CCT(nn.Module):
     """
+    Denoising convolutional transformer
     Based on the Compact Convolutional Transformer (CCT) of [Hasani et al. (2021)
         *''Escaping the Big Data Paradigm with Compact Transformers''*](
         https://arxiv.org/abs/2104.05704). It's a convolutional neural network
@@ -306,29 +338,31 @@ class CCT(nn.Module):
         minimum_cnn_out_channels=16,
         cnn_kernel_size=3,
         cnn_kernel_stride=1,
-        cnn_kernel_padding="same",
+        cnn_padding="same",
         cnn_kernel_dilation=1,
         cnn_kernel_groups=1,
-        cnn_activation: nn.Module = nn.ReLU,
+        cnn_activation: nn.Module = ReLU,
         cnn_activation_kwargs: Optional[dict] = None,
         cnn_dropout=0.0,
-        pooling_type="maxpool",
+        pooling_type="concat",  # max, average or concat
         pooling_kernel_size=3,
         pooling_kernel_stride=2,
-        pooling_kernel_padding=1,
-        transformer_position_embedding="absolute",  # absolute or relative
+        pooling_padding=1,
+        intermediate_feedforward_layer=True,
+        transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
         transformer_heads=4,
         transformer_mlp_ratio=2,
-        transformer_bos_tokens=4,
-        transformer_activation: nn.Module = nn.GELU,
+        transformer_bos_tokens=0,
+        transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
-        mlp_dropout=0.0,
-        msa_dropout=0.1,
-        stochastic_depth=0.1,
+        transformer_mlp_dropout=0.0,
+        transformer_msa_dropout=0.1,
+        transformer_stochastic_depth=0.1,
+        batch_norm_outputs=True,
+        initial_batch_norm=True,
         linear_module=nn.Linear,
-        batch_norm=True,
         image_classes=100,
     ):
 
@@ -350,13 +384,13 @@ class CCT(nn.Module):
                 "SwiGLU": SwiGLU,
             }[transformer_activation]
 
-        self.encoder = CCTEncoder(
+        self.encoder = ViTEncoder(
             input_size=input_size,
             cnn_in_channels=cnn_in_channels,
             minimum_cnn_out_channels=minimum_cnn_out_channels,
             cnn_kernel_size=cnn_kernel_size,
             cnn_kernel_stride=cnn_kernel_stride,
-            cnn_kernel_padding=cnn_kernel_padding,
+            cnn_padding=cnn_padding,
             cnn_kernel_dilation=cnn_kernel_dilation,
             cnn_kernel_groups=cnn_kernel_groups,
             cnn_activation=cnn_activation,
@@ -365,7 +399,8 @@ class CCT(nn.Module):
             pooling_type=pooling_type,
             pooling_kernel_size=pooling_kernel_size,
             pooling_kernel_stride=pooling_kernel_stride,
-            pooling_kernel_padding=pooling_kernel_padding,
+            pooling_padding=pooling_padding,
+            intermediate_feedforward_layer=intermediate_feedforward_layer,
             transformer_position_embedding=transformer_position_embedding,
             transformer_embedding_size=transformer_embedding_size,
             transformer_layers=transformer_layers,
@@ -374,14 +409,17 @@ class CCT(nn.Module):
             transformer_bos_tokens=transformer_bos_tokens,
             transformer_activation=transformer_activation,
             transformer_activation_kwargs=transformer_activation_kwargs,
-            mlp_dropout=mlp_dropout,
-            msa_dropout=msa_dropout,
-            stochastic_depth=stochastic_depth,
+            mlp_dropout=transformer_mlp_dropout,
+            msa_dropout=transformer_msa_dropout,
+            transformer_stochastic_depth=transformer_stochastic_depth,
             linear_module=linear_module,
-            batch_norm=batch_norm,
+            initial_batch_norm=initial_batch_norm,
         )
         self.pool = SequencePool(
-            transformer_embedding_size, linear_module, image_classes
+            transformer_embedding_size,
+            linear_module,
+            image_classes,
+            batch_norm=batch_norm_outputs,
         )
 
     @property
